Raw64

Struct Raw64 

pub struct Raw64 { /* private fields */ }
An internal representation of Residue64 without an associated Modulus64.

Conceptually, Residue64 = Raw64 + Modulus64. Raw64 stores the value part alone, without holding a reference to its modulus.

This separation is useful for reducing the size of collections of Residue64 and for avoiding self-referential structures when a type needs to contain both a residue and its modulus.

Implementations

Source

impl Raw64

pub const fn into_residue(self, modulus: &Modulus64) -> Residue64<'_>

Attaches a modulus and returns a Residue64.

Typically, this only needs to be called once per computation because Raw64 and Residue64 can interact.

Caution

This does not perform validation or reduction. The caller must ensure the modulus is correct for this value.