Residue32

Struct Residue32 

pub struct Residue32<'a> { /* private fields */ }
Residue with odd modulus which is no more than 2_654_435_769.

Fast modular multiplication

Residue32 provides fast modular multiplication called Plantard multiplication. This method saves one multiplication when either of two values of a multiplication is used multiple times. Therefore, Residue32::pow will be faster than that using Montgomery multiplication.

Usage

use lib_modulo::Modulus32;

// set modulus
let modulus = Modulus32::new(3);

// performs modular arithmetics
let one = modulus.residue(1);
let two = modulus.residue(2);
let five = modulus.residue(5);
assert_eq!(two * five, one)

Two residues with different modulus can interact, but the result will be meaningless. It is highly recommended to use a block to ensure that Modulus32, therefore Residue32s, are dropped.

Implementations

Source

impl<'a> Residue32<'a>

pub const fn is_zero(self) -> bool

Checks whether self is 0.

Example
use lib_modulo::Modulus32;

let modulus = Modulus32::new(5);
assert!(modulus.residue(10).is_zero())
pub const fn get(self) -> u64

Returns the residue.

Example
use lib_modulo::Modulus32;

let modulus = Modulus32::new(7);
assert_eq!(modulus.residue(10).get(), 3)
pub const fn modulus(&self) -> u64

Returns the modulus.

Example
use lib_modulo::Modulus32;

let modulus = Modulus32::new(11);
assert_eq!(modulus.residue(2).modulus(), 11);
pub const fn pow(self, exp: u32) -> Self

Raises self to the power of exp, using exponentiation by squaring.

Time complexity

Θ(log exp)

Example
use lib_modulo::Modulus32;

let modulus = Modulus32::new(1001);
let residue = modulus.residue(2);
for exp in 0..64 {
    assert_eq!(residue.pow(exp).get(), (1 << exp) % 1001)
}
pub const fn try_inv(self) -> Result<Self, u64>

Calculates the modular inverse of self, using extended binary GCD algorithm.

Modular inverse can be defined if and only if self and the modulus is coprime.

  • Ok(x) : x is the modular inverse.
  • Err(x): x is the GCD of self and the modulus, where gcd(0, a) is defined to be a.
Time complexity

O(log self)

Example
use lib_modulo::Modulus32;

let modulus = Modulus32::new(3 * 5);

let residue = modulus.residue(2);
assert!(residue.try_inv().is_ok_and(|inv| (inv * residue).get() == 1));

let residue = modulus.residue(6);
assert!(residue.try_inv().is_err_and(|gcd| gcd == 3));